#ifndef GA_H
#define GA_H

#include <iostream>
#include <vector>
#include <algorithm>
#include <math.h>

using namespace std;

template <class Gene>
struct gene_fitness_pair {
  Gene gene;
  double f;
  gene_fitness_pair(const Gene &gene, double f) : gene(gene), f(f) {}
};

//this really ought to live is some standard lib-
//ofcouse, if i define it myself, it will conflict with everyone else's...
//int min(int a, int b) { return a<b ? a : b; }


//So this is somewhere between a GA and a simulated annealer,
//there is no crossover, mutation probability, mutation magnitude, 
//and population size decrease over time.
//(I do plan to template on double as well... )
template <class Fitness,         // double foo(const Gene& v)
	      class CreateChildren,  // vector<Gene> children(vector< gene_fitness_pair >, int t )
          class Gene>            // vector<double> or vector<bool> most likely
 
struct GA {
  
  typedef gene_fitness_pair<Gene> gfp;

  const Fitness fitness;
  const CreateChildren create_children;

  int time;

  vector<Gene > population; 
  vector<gene_fitness_pair<Gene> > population_with_fitness;

  GA(Fitness f, CreateChildren cc)
    : fitness(f), create_children(cc), time(0)
  {
	  population=create_children(vector<gfp >(),time); 
  }


  struct gene_sort {
    bool operator ()(const gene_fitness_pair<Gene> &a,const gene_fitness_pair<Gene> &b) {
      return Fitness::favor_less ? a.f<b.f : a.f>b.f;
    }
  };

  void update_fitness() {
    population_with_fitness.clear();
    for(vector<Gene >::iterator i = population.begin();i!=population.end();i++) {
      population_with_fitness.push_back( gene_fitness_pair<Gene>(*i,fitness(*i)) );
    }

    sort( population_with_fitness.begin(), population_with_fitness.end(),gene_sort());
  }
  
  void breed() {
    population = create_children(population_with_fitness,time++);
  }


};
  

template< class MutationProb,    // double foo(int t) .... (maybe i should add a rank param?)
	  class MutationDelta,
	  class InitPopulation>     // double foo(int t)
struct NolfiChildren {

  const int vector_length;
  const InitPopulation init_pop;
  const MutationProb mutation_prob;
  const MutationDelta mutation_delta;
  
  NolfiChildren(int vec_len, MutationProb mp, MutationDelta md, InitPopulation init) :
    vector_length(vec_len),init_pop(init),
    mutation_prob(mp), mutation_delta(md) {}
 
  vector<double> new_child(const vector<double> &parent, int t) const {
    vector<double> child=parent;
    for(vector<double>::iterator i=child.begin();i!=child.end();i++) {
      if(mutation_prob(t)>drand48()) *i=*i+mutation_delta(t);
    }
    return child;
  }

  vector<vector<double> > operator()(const vector< gene_fitness_pair<vector<double> > > &pop, int t) const {
    if(t==0) return init_pop(vector_length,100);
    vector<vector<double> > children(100);
    
    const int nchildren=5;

    for(int i=0;i<20;i++) {
      for(int k=0;k<nchildren;k++)
	//a little hack to throw some new blood into the pop
	if(pop[i].f == 0 && (drand48() < .5))
	  children[i*nchildren+k]= init_pop(vector_length,1).back();
	else
	  children[i*nchildren+k]=new_child(pop[i].gene,t);
    }

    return children;
  }
};

//Fitness function object used to test the GA
struct CosSimFitness {
#ifdef GCC
	static const bool favor_less=false;
#else
	static const bool favor_less;
#endif

  const int VectorLength;
  const vector<double> goal;

  CosSimFitness(int len) : 
    VectorLength(len), goal(VectorLength,.5) {}
  CosSimFitness(int len, const vector<double> &goal) : 
    VectorLength(len), goal(goal) {}

  double operator()(const vector<double> &v) const {
    //Assert(v.size()==VectorLength,"goal and input vectors must match size");
    double dot=0;
    double mag1=0,mag2=0;
    for(int i=0;i<VectorLength;i++) {
      dot+=v[i]*goal[i];
      mag1+=v[i]*v[i];
      mag2+=goal[i]*goal[i];
    }
    return dot/sqrt(mag1*mag2);
  }

};

#ifndef GCC
const bool CosSimFitness::favor_less=false;
#endif
  
//Fitness function object used to test the GA
struct SSD_Fitness {
  const int VectorLength;
  const vector<double> goal;
#ifdef GCC
	static const bool favor_less=true;
#else
	static const bool favor_less;
#endif

  SSD_Fitness(int len) : 
    VectorLength(len), goal(VectorLength,.5) {}
  SSD_Fitness(int len, const vector<double> &goal) : 
    VectorLength(len), goal(goal) {}

  double operator()(const vector<double> &v) const {
    //Assert(v.size()==VectorLength,"goal and input vectors must match size");
    double ssd=0;
    for(int i=0;i<VectorLength;i++) {
      ssd+=(v[i]-goal[i])*(v[i]-goal[i]);

    }
    return ssd;
  }

};

#ifndef GCC
const bool SSD_Fitness::favor_less=true;
#endif
 
struct UniformDist {
  const double scale;
  const double displace;
  UniformDist(double s=1.0, double d=0.0) : scale(s), displace(d) {}
  
  vector<vector<double> > operator()(int vec_len, int pop_size) const {
    vector<vector<double> > v;
    v.resize( pop_size);
    for(int i=0;i<pop_size;i++) {
	v[i].resize(vec_len);
	for(int k=0;k<vec_len;k++) {
	  v[i][k]=scale*drand48()+displace;
	}
    }
    return v;
  }
};

double drand48_gaussian() {
  double x,y,q;
   do
  {
     x=2*drand48()-1;
     y=2*drand48()-1;
  } while((q=sqrt(x*x+y*y))>=1);
  return x/q*sqrt(-4*log(q));
}

struct Gaussian {
  const double scale;
  Gaussian(double scale) : scale(scale) {}
  double operator()(int t) const {
    return drand48_gaussian()*scale;
  }
};  

struct ExponentialDecay {
  const double c, k;
  ExponentialDecay(double c, double k) : c(c), k(k) {}
  double operator()(int t) const {
    return c*exp(k*t);
  }

};

template <class T>
struct Const {
  const T v;
  Const(T v) : v(v) {}
  T operator()(int t) const { return v; }
};


struct LinearDecay {
  const double r0, dr;
  LinearDecay(double r0, double dr) :
    r0(r0), dr(dr) {}
  double operator()(int t) const {
    return r0-dr*t;
  }
};



typedef GA<  SSD_Fitness,    //fitness function
	     NolfiChildren<Const<double>, Gaussian, UniformDist>,
	     vector<double> >
TestGA_P;  

struct TestGA : public TestGA_P {

  TestGA(int vec_len) : 
    TestGA_P(SSD_Fitness(vec_len), 
	     NolfiChildren<Const<double>, Gaussian, UniformDist>
						       ( vec_len,
								 Const<double>(.01), //mutation rate
								 Gaussian(5),        //mutation mag
								 UniformDist(1,0))
	    )
  {}

};
      

ostream& operator<<(ostream &out,const vector<double> &v) {
  for(int i=0;i<v.size();i++) {
    out << v[i] << " ";
  }
  out << endl;
  return out;
}


#endif
